# Copyright 2023-2024 Broadcom
# SPDX-License-Identifier: Apache-2.0
import logging


class VdkBoundLogger(logging.LoggerAdapter):
    """
    This class wraps around a Logger to append additional context to log records.
    Logger-wide context can be included by providing it in the `context` parameter of the
    `bind_logger` method.
    Log statement-specific context can be included by including an `extra` kwarg to the log call and
    providing a dictionary object (or any object which implements __getitem__ and __iter__) with the
    required additional log fields.
    It is intended to be instantiated using the `bind_logger` function and providing it with the current
    log object and the additional context which will be included in dictionary form.

    Example use::

        import logging
        from vdk.internal.builtin_plugins.config.log_config import bind_logger

        bound_logger = bind_logger(logging.getLogger(__name__), {"logger_wide_bound_key": "logger_wide_bound_value"})
        bound.warning("From vdk_initialize", extra={"first_key": "first_value"})
        bound.warning("Something else from vdk_initialize")

    Now, the LogRecord produced by the first log statement will have both
    "logger_wide_bound_key": "logger_wide_bound_value" and "first_key": "first_value" included in its __dict__,
    while the second one will only have "first_key": "first_value" in it.

    Note that these will not be automatically included in the output log and must be processed downstream by
    something like vdk-structlog.
    """

    def process(self, msg: str, kwargs: dict):
        # merge bound extra dict with existing extra dict if any
        if "extra" in kwargs:
            kwargs["extra"] = {**self.extra, **kwargs["extra"]}
        else:
            kwargs["extra"] = self.extra
        return msg, kwargs


def bind_logger(log: logging.Logger, context: dict):
    """
    Instantiates a VdkBoundLogger object.

    :param log: current log object
    :param context: dictionary containing the additional context we want to include in log statements
    :return: a VdkBoundLogger instance
    """
    return VdkBoundLogger(log, context)



# Copyright 2023-2024 Broadcom
# SPDX-License-Identifier: Apache-2.0
import logging
import pathlib
import textwrap
from typing import List
from typing import Optional

from vdk.api.job_input import IJobArguments
from vdk.api.job_input import IJobInput
from vdk.api.job_input import ITemplate
from vdk.internal.builtin_plugins.connection.impl.router import ManagedConnectionRouter
from vdk.internal.builtin_plugins.connection.managed_connection_base import (
    ManagedConnectionBase,
)
from vdk.internal.builtin_plugins.ingestion.ingester_router import IngesterRouter
from vdk.internal.builtin_plugins.job_properties.properties_router import (
    PropertiesRouter,
)
from vdk.internal.builtin_plugins.job_secrets.secrets_router import (
    SecretsRouter,
)
from vdk.internal.builtin_plugins.run.execution_results import ExecutionResult
from vdk.internal.builtin_plugins.run.sql_argument_substitutor import (
    SqlArgumentSubstitutor,
)
from vdk.internal.core.context import CoreContext
from vdk.internal.core.errors import SkipRemainingStepsException
from vdk.internal.core.errors import UserCodeError
from vdk.internal.core.statestore import CommonStoreKeys


# TODO make this extensible so new job inputs can be added by plugins
class JobInput(IJobInput):
    """
    Implementation of JobInput interface.
    It just forwards to actual implementation of a given functionality and should do nothing else.
    """

    def __init__(
        self,
        name: str,
        job_directory: pathlib.Path,
        core_context: CoreContext,
        job_arguments: IJobArguments,
        templates: ITemplate,
        managed_connection_builder: ManagedConnectionRouter,
        ingester: IngesterRouter,
        properties_router: PropertiesRouter,
        secrets_router: SecretsRouter,
    ):
        """Constructor."""

        self.__name = name
        self.__job_directory = job_directory
        self.__statestore = core_context.state
        self.__job_arguments = job_arguments
        self.__templates = templates
        self.__managed_connection_builder = managed_connection_builder
        self.__ingester = ingester
        self.__properties_router = properties_router
        self.__secrets_router = secrets_router
        self.__vdk_internal_telemetry = None

    # Connections

    def get_managed_connection(self, database: str = None) -> ManagedConnectionBase:
        if not database:
            return self.__managed_connection_builder.open_default_connection()
        return self.__managed_connection_builder.open_connection(dbtype=database)

    def get_arguments(self) -> dict:
        return self.__job_arguments.get_arguments()

    # Properties

    def get_property(self, name, default_value=None):
        return self.__properties_router.get_property(name, default_value)

    def get_all_properties(self):
        return self.__properties_router.get_all_properties()

    def set_all_properties(self, properties):
        return self.__properties_router.set_all_properties(properties)

    def get_secret(self, name, default_value=None):
        return self.__secrets_router.get_secret(name, default_value)

    def get_all_secrets(self):
        return self.__secrets_router.get_all_secrets()

    def set_all_secrets(self, secrets):
        return self.__secrets_router.set_all_secrets(secrets)

    def _substitute_query_params(self, sql: str):
        sql = textwrap.dedent(sql).strip("\n") + "\n"
        query = sql
        sql_susbstitute_args = {}
        if not self.__properties_router.has_properties_impl():
            logging.getLogger(__name__).info(
                "Data Job Properties has not been initialized., "
                "so I won't be able to provide query properties substitution capabilities from job properties."
                "If passed job arguments will still be used"
            )
        else:
            sql_susbstitute_args.update(self.get_all_properties())
        sql_susbstitute_args.update(self.get_execution_properties())

        sql_args = self.get_arguments()
        if not sql_args or type(sql_args) != dict:
            logging.getLogger(__name__).debug(
                "No arguments are passed for Data Job, "
                "so I won't be able to provide query parameter substitution capabilities with job arguments."
            )
        else:
            sql_susbstitute_args.update(sql_args)
        query = SqlArgumentSubstitutor(sql_susbstitute_args).substitute(query)

        return query

    def execute_query(self, sql: str, database: str = None):
        if not sql or not sql.strip():
            raise UserCodeError("Trying to execute an empty SQL query.")

        query = self._substitute_query_params(sql)

        connection = self.get_managed_connection(database)
        return connection.execute_query(query)

    def send_object_for_ingestion(
        self,
        payload: dict,
        destination_table: Optional[str] = None,
        method: Optional[str] = None,
        target: Optional[str] = None,
        collection_id=None,
    ):
        self.__ingester.send_object_for_ingestion(
            payload,
            destination_table,
            method,
            target,
            collection_id,
        )

    def send_tabular_data_for_ingestion(
        self,
        rows: iter,
        column_names: List,
        destination_table: Optional[str] = None,
        method: Optional[str] = None,
        target: Optional[str] = None,
        collection_id: Optional[str] = None,
    ):
        self.__ingester.send_tabular_data_for_ingestion(
            rows,
            column_names,
            destination_table,
            method,
            target,
            collection_id,
        )

    def execute_template(
        self, template_name: str, template_args: dict, database: str = "default"
    ) -> ExecutionResult:
        if self.__templates:
            result = self.__templates.execute_template(
                template_name, template_args, database
            )

            return result
        else:
            raise NotImplemented("Templates not wired to JobInput")

    def get_name(self) -> str:
        return self.__name

    def get_job_directory(self) -> pathlib.Path:
        return self.__job_directory

    def get_execution_properties(self) -> dict:
        start_time = self.__statestore.get(CommonStoreKeys.START_TIME)
        return {
            "pa__execution_id": self.__statestore.get(CommonStoreKeys.EXECUTION_ID),
            "pa__job_start_unixtime": str(int(start_time.timestamp())),
            "pa__job_start_ts_expr": f"cast ({start_time.timestamp()} as timestamp)",
            "pa__op_id": self.__statestore.get(CommonStoreKeys.OP_ID),
        }

    def skip_remaining_steps(self) -> None:
        error_message_lines = [
            "Job/template execution was skipped.",
            "Job/template execution was skipped from job/template step code.",
            "Job/template called the job_input.skip_remaining_steps() method.",
            "The remaining steps (if any) will not be executed and current job/template execution "
            + "will finish. The job/template will terminate with a success status.",
            "Revise job/template code and determine need for skipping. "
            + "If cancellation behaviour no longer desired, refactor the job/template code.",
        ]
        raise SkipRemainingStepsException(*error_message_lines)

    def get_temporary_write_directory(self) -> pathlib.Path:
        path_string = self.__statestore.get(CommonStoreKeys.TEMPORARY_WRITE_DIRECTORY)
        return pathlib.Path(path_string)



# Copyright 2023-2024 Broadcom
# SPDX-License-Identifier: Apache-2.0
# Copyright 2021-2024 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging
import os
from typing import Dict

from vdk.api.plugin.hook_markers import hookimpl
from vdk.internal.builtin_plugins.run.job_context import JobContext

VDK_ = "VDK_"


log = logging.getLogger(__name__)


class SecretsConfigPlugin:
    def __normalize_env_key(self, key):
        return key.replace("-", "_").replace(".", "_").upper()

    def _get_env_var(self, key, env: Dict[str, str]):
        normalized_key = self.__normalize_env_key(key)
        normalized_key_with_vdk_prefix = VDK_ + normalized_key
        value = (
            env.get(normalized_key_with_vdk_prefix)
            if normalized_key_with_vdk_prefix in env
            else env.get(normalized_key)
        )
        return value

    @hookimpl(trylast=True)
    def initialize_job(self, context: JobContext):
        upper_cased_env = {k.upper(): v for k, v in os.environ.items()}
        secrets = context.job_input.get_all_secrets()
        for key, value in secrets.items():
            env_var = self._get_env_var(key, upper_cased_env)
            # override only if there is no corresponding environment variable
            overridden_in_section = False
            if env_var is None:
                sections = [
                    section
                    for section in context.core_context.configuration.list_sections()
                    if section != "vdk"
                ]
                # try to match the secret to a specific config section and override it
                for section in sections:
                    if section.lower() in key.lower():
                        log.info(f"Overriding config {key} with secret")
                        context.core_context.configuration.override_value(
                            key, value, section
                        )
                        overridden_in_section = True
                        break

                # if it wasn't overridden in a section, override it in the main section
                if not overridden_in_section:
                    context.core_context.configuration.override_value(key, value)



# Copyright 2023-2024 Broadcom
# SPDX-License-Identifier: Apache-2.0
from typing import Container
from typing import Optional

from vdk.internal.builtin_plugins.connection.decoration_cursor import DecorationCursor
from vdk.internal.builtin_plugins.connection.execution_cursor import (
    ExecuteOperationResult,
)
from vdk.internal.builtin_plugins.connection.execution_cursor import ExecutionCursor
from vdk.internal.builtin_plugins.connection.recovery_cursor import RecoveryCursor


class IDatabaseManagedConnection:
    def db_connection_validate_operation(
        self, operation: str, parameters: Optional[Container]
    ) -> None:
        """
        Validates the query and parameters.

        For example:
        db_connection_validate_operation(operation, parameters):
            if(_max_query_length_limit_exceeded(operation, parameters)):
                raise Exception("max query length limit exceeded")

        :param operation: str
            Database operation - a SQL query or command to be executed.
        :param parameters: Optional[Container]
            Parameters may be provided as sequence or mapping and will be bound to variables in the operation.
            Variables are specified in a database-specific notation. See chosen database documentation for details.
        :return:
        """
        pass

    def db_connection_decorate_operation(
        self, decoration_cursor: DecorationCursor
    ) -> None:
        """
        Curates the operation and parameters.
        If an exception is raised, the cursor execution will be suspended.

        For example:
        db_connection_decorate_operation(decoration_cursor: DecorationCursor):
            managed_operation = decoration_cursor.get_managed_operation()
            managed_operation.set_operation("prefix" +
                                            managed_operation.get_operation())

        :param decoration_cursor: DecorationCursor
            A PEP249Cursor implementation purposed for query and parameters decoration.
            Provides operation details and tooling.
        :return:
        """
        pass

    def db_connection_execute_operation(
        self, execution_cursor: ExecutionCursor
    ) -> Optional[ExecuteOperationResult]:
        """
        The method that executes the actual SQL query using execution cursor.
        For example: let's say we are writing vdk-impala plugin and want to print more debug info
            which is available from the Impala native cursor (provided by impyla library)
                db_connection_execute_operation(execution_cursor: ExecutionCursor) -> Optional[int]:
                    yield # let the query execute first
                    c = cast(impala.interface.Cursor, execution_cursor)
                    log.info(f"Query {execution_cursor.get_managed_operation().get_operation()} debug info:"
                            f"summary: {c.get_summary()}, profile: {c.get_profile()}")
        :param execution_cursor: ExecutionCursor
        A PEP249Cursor implementation purposed for actual query execution.
        """
        managed_operation = execution_cursor.get_managed_operation()
        if managed_operation.get_parameters():
            native_result = execution_cursor.execute(
                managed_operation.get_operation(), managed_operation.get_parameters()
            )
        else:
            native_result = execution_cursor.execute(managed_operation.get_operation())
        return ExecuteOperationResult(native_result)

    def db_connection_recover_operation(self, recovery_cursor: RecoveryCursor) -> None:
        """
        Recovers the operation initiated. Retries made number is auto-incremented.
        If no exception is raised, cursor is considered successfully executed.

        For example:
        db_connection_recover_operation(recovery_cursor: RecoveryCursor):
            while recovery_cursor.get_retries() < MAX_RETRIES:
                try:
                    recovery_cursor.execute("helper query")
                    recovery_cursor.retry_operation()
                    return
                except:
                    time.sleep(3 * recovery_cursor.get_retries()) # backoff
            raise recovery_cursor.get_exception()

        :param recovery_cursor: RecoveryCursor
            A PEP249Cursor implementation purposed for query and parameters recovery.
            Provides operation details and tooling.
        :return:
        """
        raise recovery_cursor.get_exception()

    def db_connection_after_operation(self, execution_cursor: ExecutionCursor) -> None:
        """
        This method runs after the execution of the database operation.
        It can be used for tasks such as auditing, logging, cleanup, or additional validations.

        For example:
        db_connection_after_operation(post_operation_cursor):
             hive_cursor = cast(HiveServer2Cursor, execution_cursor)
             lineage_data = self._get_lineage_data(hive_cursor)
             if (lineage_data):
                    lineage_logger.send(lineage_data)

        :param execution_cursor: ExecutionCursor
        :return:
        """
        pass



# Copyright 2023-2024 Broadcom
# SPDX-License-Identifier: Apache-2.0
from typing import Any
from typing import Optional

from vdk.internal.core.errors import BaseVdkError
from vdk.internal.core.errors import ResolvableBy


class IngestionException(BaseVdkError):
    """
    Base Exception for all custom exceptions related to the ingestion process.
    This is intended to catch general exceptions that do not fit into more specific categories.
    """

    def __init__(self, message: str, resolvable_by: Optional[ResolvableBy] = None):
        super().__init__(None, resolvable_by, message)


class PayloadIngestionException(IngestionException):
    """
    Base Exception for all payload-related issues during the ingestion process.
    """

    def __init__(
        self,
        payload_id: str,
        message: str,
        destination_table: str = "",
        target: str = "",
        resolvable_by: Optional[ResolvableBy] = None,
    ):
        """
        :param payload_id: ID of the payload. This is a way for user to find out which payload failed to ingest.
                           Left empty if such information is not available
        """
        super().__init__(message=message, resolvable_by=resolvable_by)
        # we are purposefully are not putting payload id
        # in the message, so it's not logged in case it's sensitive
        self.payload_id = payload_id
        self.destination_table = destination_table
        self.target = target


class EmptyPayloadIngestionException(PayloadIngestionException):
    """
    Raised when an empty payload is encountered during ingestion and it is not expected.
    """

    def __init__(
        self,
        message: Optional[str] = None,
        resolvable_by: Optional[ResolvableBy] = None,
    ):
        if not message:
            message = "Payload given to ingestion method should not be empty."
        super().__init__(payload_id="", message=message, resolvable_by=resolvable_by)


class InvalidPayloadTypeIngestionException(PayloadIngestionException):
    """
    Raised when the payload provided for ingestion has an invalid type.
    """

    def __init__(
        self,
        payload_id: str,
        expected_type: str,
        actual_type: str,
        message: Optional[str] = None,
        resolvable_by: Optional[ResolvableBy] = None,
    ):
        """
        :param expected_type: The expected type for the payload
        :param actual_type: The actual type of the payload
        """
        if not message:
            message = "Invalid ingestion payload type."
        super().__init__(
            message=f"{message} Expected type: {expected_type}, actual type: {actual_type}.",
            payload_id=payload_id,
            resolvable_by=resolvable_by,
        )


class JsonSerializationIngestionException(PayloadIngestionException):
    """
    Raised when a payload is not JSON-serializable during ingestion.
    """

    def __init__(
        self,
        payload_id: str,
        original_exception: Exception,
        message: str = "",
        resolvable_by: Optional[ResolvableBy] = None,
    ):
        """
        :param original_exception: The original exception triggering this error
        """
        if not message:
            message = "Payload is not json serializable."
        super().__init__(
            message=f"{message} Failure caused by {original_exception}",
            payload_id=payload_id,
            resolvable_by=resolvable_by,
        )


class InvalidArgumentsIngestionException(IngestionException):
    """
    Raised when an argument provided to a data ingestion method does not meet the expected constraints.
    """

    def __init__(
        self,
        param_name: str,
        param_constraint: str,
        actual_value: Any,
        message: str = "",
        resolvable_by: ResolvableBy = None,
    ):
        """
        :param param_name: The name of the parameter that caused the exception.
        :param param_constraint: Description of the constraint that the parameter failed to meet.
        :param actual_value: The actual value that was passed for the parameter.
        """
        super().__init__(
            message=f"Ingestion parameter '{param_name}' is not valid. "
            f"It must match constraint {param_constraint} but was '{actual_value}'. "
            f"{message}",
            resolvable_by=resolvable_by,
        )
        self.param_name = param_name
        self.param_constraint = param_constraint


class PreProcessPayloadIngestionException(PayloadIngestionException):
    """
    Raised when an error occurs during the pre-processing phase of payload ingestion.
    This is specifically when plugin hook pre_ingest_process raises an exception.
    """

    def __init__(
        self,
        payload_id: str,
        destination_table: str,
        target: str,
        message: str,
        resolvable_by: ResolvableBy = None,
    ):
        super().__init__(
            message=message,
            destination_table=destination_table,
            target=target,
            payload_id=payload_id,
            resolvable_by=resolvable_by,
        )


class PostProcessPayloadIngestionException(PayloadIngestionException):
    """
    Raised when an error occurs during the post-processing phase of payload ingestion.
    This is specifically when plugin hook post_ingest_process raises an exception.
    """

    def __init__(
        self,
        payload_id: str,
        destination_table: str,
        target: str,
        message: str,
        resolvable_by: ResolvableBy = None,
    ):
        super().__init__(
            message=message,
            destination_table=destination_table,
            target=target,
            payload_id=payload_id,
            resolvable_by=resolvable_by,
        )



# Copyright 2023-2024 Broadcom
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import logging
from dataclasses import dataclass
from dataclasses import field
from typing import Any

from vdk.internal.core.errors import VdkConfigurationError

log = logging.getLogger(__name__)


def convert_value_to_type_of_default_type(key: str, v: Any, default_value: Any) -> Any:
    """
    Converts a given configuration value to the type of its default value if a default is provided.

    Args:
        key (str): The key for the configuration setting.
        v (Any): The value to be converted.
        default_value (Any): The default value which defines the type to which the value should be converted.

    Returns:
        Any: The value converted to the type of the default value.

    Raises:
        VdkConfigurationError: If the value cannot be converted to the required type or if an invalid value is provided for a boolean.
    """
    # Don't bother converting NoneType
    if v is None:
        return None
    if default_value is not None:
        if isinstance(default_value, bool):
            allowed_values = ["true", "false", "1", "0", "yes", "no", "y", "n"]
            v_str = str(v).lower()
            if v_str not in allowed_values:
                raise VdkConfigurationError(
                    f"Provided configuration '{key}={v}' is invalid. Allowed values for {key} are {allowed_values}"
                )
            return v_str in ["true", "1", "yes", "y"]
        else:
            try:
                return type(default_value)(v)
            except ValueError:
                raise VdkConfigurationError(
                    f'Provided configuration "{key}={v}" is invalid. Cannot cast "{v}" to {type(default_value).__name__}'
                )
    return v


def _normalize_config_string(key: str):
    return str(key).lower()


@dataclass
class ConfigEntry:
    """
    A dataclass to store individual configuration item properties.

    Attributes:
        value: The current value of the configuration item.
        default: The default value of the configuration item.
        description: A string describing the configuration item.
        sensitive: A boolean indicating if the configuration item is sensitive.
    """

    value: Any = None
    default: Any = None
    description: str | None = None
    sensitive: bool = False


@dataclass(frozen=True)
class Configuration:
    """
    Immutable configuration organized by sections, supporting optional section specification for key searches
    in non-prefixed sections.

    Each section contains configuration entries defined by the ConfigEntry dataclass, enabling complex
    configurations setups with optional simpler access patterns when section prefixes are well-defined.

    Attributes:
        _sections (Dict[str, Dict[str, ConfigEntry]]): A dictionary where each key represents a section name
        and each value is a dictionary of configuration keys to ConfigEntry instances.
    """

    def __init__(self, sections: dict[str, dict[str, ConfigEntry]]):
        if sections:
            if "vdk" not in sections:
                sections["vdk"] = {}
            object.__setattr__(self, "_sections", sections)
        else:
            object.__setattr__(self, "_sections", {"vdk": {}})

    def get_value(self, key: str, section: str | None = None) -> Any | None:
        """
        Retrieve the current value of a configuration key, optionally searching only within a specific section or
        across all non-prefixed sections if no section is provided.

        Args:
            key (str): The configuration key to search for.
            section (Optional[str]): The specific section to search within. If None, searches across all sections
            not starting with 'vdk_'. This is kept for backwards compatibility.

        Returns:
            Optional[Any]: The current value of the configuration key if found, otherwise None.
        """
        key = _normalize_config_string(key).replace(f"{section}_", "")
        if section is None:
            if key in self._sections["vdk"]:
                entry = self._sections.get("vdk", {}).get(key, ConfigEntry())
                return entry.default if entry.value is None else entry.value
            return None
        else:
            section = _normalize_config_string(section)
            entry = self._sections.get(section, {}).get(key, ConfigEntry())
            return entry.default if entry.value is None else entry.value

    def override_value(self, key: str, value: Any, section: str = None):
        key = _normalize_config_string(key).replace(f"{section}_", "")
        if section is None:
            if key in self._sections["vdk"]:
                self._sections["vdk"][key].value = value
        else:
            section = _normalize_config_string(section)
            if section in self._sections and key in self._sections[section].keys():
                self._sections[section][key].value = value
            else:
                raise VdkConfigurationError(
                    "The value you are trying to override is not existing. "
                    f"Check the {section} section and {key} key, again!"
                )

    def get_required_value(self, key: str, section: str | None = None) -> Any:
        """
        Retrieve the required value of a configuration key, optionally from a specific section or
        across all non-prefixed sections, raising an error if the key is not found.

        Args:
            key (str): The configuration key.
            section (Optional[str]): The section from which to retrieve the value. If None, searches across
            all non-'vdk_' prefixed sections. This check is kept for backwards compatibility.

        Returns:
            Any: The value of the configuration key.

        Raises:
            ValueError: If the configuration key is missing from the specified section or non-prefixed sections.
        """
        value = self.get_value(key, section)
        if value is None:
            if section:
                raise VdkConfigurationError(
                    f"Required configuration {key} in section {section} is missing."
                )
            else:
                raise VdkConfigurationError(
                    f"Required configuration {key} is missing in non-vdk prefixed sections."
                )
        return value

    def get_description(self, key: str, section: str | None = None) -> str | None:
        """
        Retrieve the description of a configuration key, optionally from a specific section or
        across all non-prefixed sections.

        Args:
            key (str): The configuration key.
            section (Optional[str]): The section to search within.
            If None, searches across all non-'vdk_' prefixed sections. This check is kept for backwards compatibility.

        Returns:
            Optional[str]: The description of the configuration key if it exists, otherwise None.
        """
        key = _normalize_config_string(key).replace(f"{section}_", "")
        if section is None:
            if key in self._sections["vdk"]:
                return self._sections["vdk"].get(key, ConfigEntry()).description
            else:
                return None
        else:
            section = _normalize_config_string(section)
            return self._sections.get(section, {}).get(key, ConfigEntry()).description

    def is_sensitive(self, key: str, section: str | None = None) -> bool:
        """
        Check if a configuration key within a specified section or across all non-prefixed sections
        is marked as sensitive.

        Args:
            key (str): The configuration key.
            section (Optional[str]): The section to search within.
            If None, searches across all non-'vdk_' prefixed sections. This check is kept for backwards compatibility.

        Returns:
            bool: True if the configuration key is marked as sensitive, False otherwise.
        """
        key = _normalize_config_string(key).replace(f"{section}_", "")
        if section is None:
            if key in self._sections["vdk"]:
                return self._sections["vdk"][key].sensitive
            return False
        else:
            section = _normalize_config_string(section)
            return self._sections.get(section, {}).get(key, ConfigEntry()).sensitive

    def is_default(self, key: str, section: str | None = None) -> bool:
        """
        Return True if the configuration value for a given key uses the default value.
        If set_value is called for a specific key, this will return False, even if value == default_value.

        Args:
            key (str): The configuration key to check.
            section (Optional[str]): The section to check within, if None checks across all sections.

        Returns:
            bool: True if the value is the default value and not explicitly set, False otherwise.
        """

        key = _normalize_config_string(key).replace(f"{section}_", "")
        if section:
            section = _normalize_config_string(section)
            entries = self._sections.get(section, {})
        else:
            entries = self._sections.get("vdk", {})

        entry = entries.get(key)
        return entry is not None and entry.value is None

    def list_sections(self) -> list[str]:
        """
        List all section names in the configuration.

        Returns:
            List[str]: A list of all section names currently defined in the configuration.
        """
        return list(self._sections.keys())

    def list_keys_in_section(self, section: str) -> list[str]:
        """
        List all keys within a specified section of the configuration.

        Args:
            section (str): The name of the section from which to list keys.

        Returns:
            List[str]: A list of all configuration keys within the specified section.
             Returns an empty list if the section does not exist.

        Raises:
            ValueError: If the specified section does not exist in the configuration.
        """
        section = _normalize_config_string(section)
        if section in self._sections:
            return list(self._sections[section].keys())
        raise ValueError(f"Section '{section}' does not exist in the configuration.")

    def list_config_keys_from_main_sections(self) -> list[str]:
        """
        List all configuration keys from sections that do not start with 'vdk_'.
        The vdk_ are considered subsections of vdk.

        Returns:
            List[str]: A list of all configuration keys in non-'vdk_' prefixed sections.
        """
        keys = []
        for section, entries in self._sections.items():
            if not section.startswith("vdk_"):
                keys.extend(entries.keys())
        return list(keys)

    def list_config_keys(self) -> list[str]:
        # this is used in many plugins to be removed after the change in plugins is done
        return self.list_config_keys_from_main_sections()

    def __getitem__(self, key: str | tuple[str, str | None]) -> Any:
        """
        Allows the Configuration object to be accessed using subscript notation. Supports accessing
        with a single key or a tuple containing section and key.

        Args:
            key (Union[str, Tuple[str, Optional[str]]]): The key or a tuple with section and key to look up.

        Returns:
            Any: The value associated with the key.

        Raises:
            KeyError: If the key or section does not exist.
        """
        if isinstance(key, tuple):
            section, key = key
        else:
            section = None

        return self.get_value(section=section, key=key)


@dataclass
class ConfigurationBuilder:
    """
    A builder class for creating a Configuration object, structured to allow detailed control
    over configuration data organized by sections and keys.

    This builder supports defining complex configurations that can handle multiple sections,
    with each section potentially containing overlapping key names but with distinct settings.

    Attributes:
        __sections (Dict[str, Dict[str, ConfigEntry]]): Stores all configuration data during the building process.
                                                        Each section maps configuration keys to their respective ConfigEntry.
    """

    def __init__(self):
        self.__sections = {"vdk": {}}

    def add(
        self,
        key: str,
        default_value: Any | None = None,
        show_default_value: bool = True,
        description: str | None = None,
        is_sensitive: bool = False,
        section: str | None = "vdk",
    ) -> ConfigurationBuilder:
        """
        Adds a configuration key with its associated properties to a specified section. If the section does not
        already exist, it is created. If no section is specified, the configuration is added to the 'vdk' section.

        Args:
            key (str): The configuration key to add.
            default_value (Optional[Any]): The default value for the configuration key, used if no value is explicitly set.
            description (Optional[str]): A human-readable description of the configuration key.
            is_sensitive (bool): Indicates whether the configuration key contains sensitive information.
            show_default_value (bool): If true, the default value is included in the description of the configuration.
            section (Optional[str]): The name of the section to which the configuration key will be added.

        Returns:
            ConfigurationBuilder: Returns the builder instance to enable method chaining.
        """
        section = _normalize_config_string(section) if section else "vdk"
        key = _normalize_config_string(key)
        if not description:
            description = "No description provided."
        if (
            section in self.__sections
            and key in self.__sections[section].keys()
            and self.__sections[section][key].value
            != self.__sections[section][key].default
        ):
            preserved_value = convert_value_to_type_of_default_type(
                key=key,
                v=self.__sections[section][key].value,
                default_value=default_value,
            )
            self.__add_public(
                section,
                key,
                default_value,
                None,
                description,
                is_sensitive,
                show_default_value,
            )
            self.set_value(key=key, value=preserved_value, section=section)
        else:
            self.__add_public(
                section,
                key,
                default_value,
                None,
                description,
                is_sensitive,
                show_default_value,
            )
        return self

    def __add_public(
        self,
        section: str,
        key: str,
        default_value: Any | None,
        value: Any | None,
        description: str | None,
        is_sensitive: bool = False,
        show_default_value: bool = True,
    ) -> None:
        """
        Handles the internal logic of adding a new configuration key to a section, including managing descriptions
        and sensitivity settings. This method encapsulates additional formatting and initialization operations
        that are abstracted from the user.

        Args:
            section (str): The name of the section where the key will be added.
            key (str): The configuration key to be added.
            default_value (Optional[Any]): The default value for the configuration key.
            description (Optional[str]): The description of the configuration key; enhances user documentation.
            is_sensitive (bool): Specifies if the configuration key is sensitive.
            show_default_value (bool): Determines if the default value should be included in the description.

        Raises:
            ValueError: If the key already exists and the function is intended to only handle new additions.
        """
        if not isinstance(description, str):
            log.warning(
                f"Description for key {key} is not of type string. Converting to type string."
            )
            description = str(description)

        formatted_description = description
        if is_sensitive:
            formatted_description += " This option is marked as sensitive."
        if show_default_value and default_value is not None:
            formatted_description += f" Default value: {default_value}."

        if section not in self.__sections:
            self.__sections[section] = {}

        if not value:
            value = default_value

        if value == default_value:
            self.__sections[section][key] = ConfigEntry(
                value=None,
                default=default_value,
                description=formatted_description,
                sensitive=is_sensitive,
            )
        else:
            self.__sections[section][key] = ConfigEntry(
                value=value,
                default=default_value,
                description=formatted_description,
                sensitive=is_sensitive,
            )

    def set_value(
        self, key: str, value: Any, section: str | None = "vdk"
    ) -> ConfigurationBuilder:
        """
        Sets or updates the value for a specific configuration key within a designated section.

        Args:
            key (str): The configuration key whose value is to be updated.
            section (Optional[str]): The name of the section containing the configuration key. Defaults to 'vdk'.
            value (Any): The new value to assign to the configuration key.

        Returns:
            ConfigurationBuilder: Returns the builder instance to enable method chaining.

        Raises:
            ValueError: If the key does not exist in the given section, indicating it must be added first.
        """
        section = _normalize_config_string(section)
        key = _normalize_config_string(key)
        if section not in self.__sections or key not in self.__sections[section]:
            self.__add_public(
                section=section,
                key=key,
                default_value=None,  # it should be = value
                value=value,
                description="",
            )

        if self.__sections[section][key].default is not None:
            self.__sections[section][key].value = convert_value_to_type_of_default_type(
                key=key, v=value, default_value=self.__sections[section][key].default
            )
        else:
            self.__sections[section][key].value = value
        return self

    def list_config_keys_from_main_sections(self) -> list[str]:
        """
        List all configuration keys from sections that do not start with 'vdk_'.
        The vdk_ are considered subsections of vdk.

        Returns:
            List[str]: A list of all configuration keys in non-'vdk_' prefixed sections.
        """
        return list(self.__sections["vdk"].keys())

    def list_config_keys_for_section(self, section: str) -> list[str]:
        """
        List all keys within a specified section of the configuration.

        Args:
            section (str): The name of the section from which to list keys.

        Returns:
            List[str]: A list of all configuration keys within the specified section.
            Raises ValueError if the section does not exist.
        """
        section = _normalize_config_string(section)
        if section in self.__sections:
            return list(self.__sections[section].keys())
        else:
            raise ValueError(f"Section '{section}' does not exist.")

    def list_all_section_names(self) -> list[str]:
        """
        List all section names in the configuration.

        Returns:
            List[str]: A list of all section names currently defined in the configuration.
        """
        return list(self.__sections.keys())

    def build(self) -> Configuration:
        """
        Finalizes the building process and constructs an immutable Configuration object from the accumulated data.

        Returns:
            Configuration: The constructed immutable configuration object, ready to be used within the application.
        """
        return Configuration(sections=self.__sections)




RELEASE_VERSION='0.3.1422958728'
BUILD_DATE='Thu Aug 22 09:18:07 UTC 2024'
BUILD_MACHINE_INFO='Linux runner-jhcjxvh-project-28359933-concurrent-0 5.15.154+ #1 SMP Sat May 4 12:14:42 UTC 2024 x86_64 GNU/Linux'
GITLAB_CI_JOB_ID='7640412494'
GIT_COMMIT_SHA='7c2bd451320b905d7366196067fcfb13a719b4de'
GIT_BRANCH='main'



# Copyright 2023-2024 Broadcom
# SPDX-License-Identifier: Apache-2.0
import copy
import logging
import pathlib
from typing import Dict
from typing import Optional

from vdk.api.job_input import ITemplate
from vdk.api.plugin.plugin_input import ITemplateRegistry
from vdk.internal.builtin_plugins.run.data_job import DataJobFactory
from vdk.internal.builtin_plugins.run.execution_results import ExecutionResult
from vdk.internal.core import errors
from vdk.internal.core.context import CoreContext

log = logging.getLogger(__name__)


class TemplatesImpl(ITemplateRegistry, ITemplate):
    def __init__(
        self,
        job_name: str,
        core_context: CoreContext,
        datajob_factory: DataJobFactory = None,
        template_name: Optional[str] = None,
    ):
        self._job_name = job_name
        self._core_context = core_context
        self._registered_templates: Dict[str, Dict[str, pathlib.Path]] = {}
        self._datajob_factory = (
            DataJobFactory() if datajob_factory is None else datajob_factory
        )
        self._template_name = template_name

    def add_template(
        self, name: str, template_directory: pathlib.Path, database: str = "default"
    ):
        if (
            database in self._registered_templates
            and name in self._registered_templates[database]
            and self._registered_templates[database][name] != template_directory
        ):
            log.warning(
                f"Template with name {name} in database {database} has been registered with directory "
                f"{self._registered_templates[database][name]}."
                f"We will overwrite it with new directory {template_directory} now."
            )
        if database not in self._registered_templates:
            self._registered_templates[database] = {}
        self._registered_templates[database][name] = template_directory

    # Added database argument to support multiple DB type
    def execute_template(
        self, name: str, template_args: dict, database: str = "default"
    ) -> ExecutionResult:  # input dict immutable?
        log.info(f"Execute template {database} {name} {template_args}")
        template_directory = self.get_template_directory(name, database)
        if database != "default":
            core_context = CoreContext(
                self._core_context.plugin_registry,
                copy.deepcopy(self._core_context.configuration),
                self._core_context.state.clone(),
            )
            core_context.configuration.override_value(
                "DB_DEFAULT_TYPE", database, "vdk"
            )
            template_job = self._datajob_factory.new_datajob(
                template_directory, core_context, name=self._job_name
            )
        else:
            template_job = self._datajob_factory.new_datajob(
                template_directory, self._core_context, name=self._job_name
            )

        result = template_job.run(template_args, name)
        if result.is_failed():
            if result.get_exception_to_raise():
                raise result.get_exception_to_raise()
            else:
                errors.report_and_throw(
                    errors.PlatformServiceError(
                        f"Template `{name}` failed.",
                        f"No exception is reported. This is not expected. "
                        f"Execution steps of templates were {result.steps_list}",
                        "We will raise an exception now. Likely the job will fail.",
                        f"Check out the error and fix the template invocation "
                        f"or fix the template by installing the correct plugin."
                        f" Or open an issue on the support team or "
                        f"Versatile Data Kit or the plugin provider of the template",
                    )
                )
        return result

    def get_template_directory(
        self, name: str, database: str = "default"
    ) -> pathlib.Path:
        if (
            database in self._registered_templates
            and name in self._registered_templates[database]
        ):
            return self._registered_templates[database][name]
        else:
            errors.report_and_throw(
                errors.UserCodeError(
                    f"No registered template with name: {name} and database {database}.",
                    "Template with that name has not been registered",
                    "Make sure you have not misspelled the name of the template "
                    "or the plugin(s) providing the template is installed. "
                    f"Current list of templated is: {list(self._registered_templates.keys())}",
                )
            )
